/* 
   gencomp -- software for comparing transcript structures along a genome
   Copyright (C) 2010  Liliana Florea
   Report bugs or questions: florea@umiacs.umd.edu
*/

#include "libc.h"
#include "lib.h"
#include "compare.h"


/* Load one axis at a time */

#define DEFAULT_MATCH_RANGE      0 
#define DEFAULT_MATCH_EXON       1
#define DEFAULT_MATCH_GRAPH      0
#define DEFAULT_MATCH_STRAND     0
#define DEFAULT_MATCH_SPLICE     0

#define DEFAULT_VI               10
#define DEFAULT_VM               10

#define DEFAULT_COMPATIBILITY    STRICT


static char Usage[] = 

"%s -t file1 -t file2 [-r] [-x] [-o] [-h] [-j] [-Vi vi] [-Vm vm] -[s|w]\n\
    file1,2 transcripts or genes in ASV format\n\
    -r      compare only if genomic range matches (off)\n\
    -x      compare only if exon overlap (on)\n\
    -o      compare only if same strand (off)\n\
    -h      compare only if same graph/gene (off)\n\
    -j      compare only if common splice (off)\n\
    -s      strict compatibility = both exons and introns (default)\n\
    -w      relaxed compatibility = introns only\n\
    -Vi     allowed variation for internal boundaries (default 10)\n\
    -Vm     allowed variation for marginal boundaries (default 10)\n";

static char   errmsg[100];


static int    readGraph(FILE *,ASV_t **,int *,int *);
static int    loadASV(FILE *,ASV_t **,int *,int *,char *);
static int    line2ASV(char *buf, ASV_t *p);
static void   compare_asvs_sets(ASV_t *,int,ASV_t *,int,int,int,int,int,int,int,int,int);

static int    common_splice(ASV_t *,ASV_t *);
static void   addMatches(ASV_t *,int,ASV_t **,int *,int *);
static char  *getLine(char **,int *,FILE *);
static int    exon_overlaps(ASV_t *,ASV_t *);

int main(int argc, char *argv[])
{
   FILE    *fp1, *fp2;
   int      i, lastGAid;

   ASV_t   *asv1, *asv2;
   int      numasv1, numasv2, maxasv1, maxasv2; 

   ASV_t   *ASVs1, *ASVs2;
   int      numASVs1, maxASVs1, numASVs2, maxASVs2;
   int      st1, st2;

   int      argcount; 
   int      matchRange, matchExon, matchGraph, matchStrand, matchSplice;
   int      Vi, Vm, mode, modeset;

   matchRange  = DEFAULT_MATCH_RANGE;
   matchExon   = DEFAULT_MATCH_EXON;
   matchGraph  = DEFAULT_MATCH_GRAPH;
   matchStrand = DEFAULT_MATCH_STRAND;
   matchSplice = DEFAULT_MATCH_SPLICE;

   Vi = DEFAULT_VI;
   Vm = DEFAULT_VM;

   mode = DEFAULT_COMPATIBILITY; 

   if (argc<5) fatalf(Usage, argv[0]);

   if (strcmp(argv[1],"-t") || strcmp(argv[3],"-t"))
       fatalf(Usage, argv[0]);

   modeset = 0;
   
   argcount = 5;
   while (argcount<argc) {
      if (!strcmp(argv[argcount],"-Vi"))
         Vi = atoi(argv[++argcount]);
      else if (!strcmp(argv[argcount],"-Vm"))
         Vm = atoi(argv[++argcount]);
      else if (!strcmp(argv[argcount],"-w")) {
         if (modeset && (mode==STRICT))
            fatal("Mode already set to strict inclusion.\n");
         mode = WEAK;
         modeset = 1;
      } else if (!strcmp(argv[argcount],"-s")) {
         if (modeset && (mode==WEAK))
            fatal("Mode already set to weak inclusion.\n");
         mode = STRICT;
         modeset = 1;
      } else if (!strncmp(argv[argcount],"-r",2) ||
                 !strncmp(argv[argcount],"-x",2) ||
                 !strncmp(argv[argcount],"-o",2) ||
                 !strncmp(argv[argcount],"-h",2) ||
                 !strncmp(argv[argcount],"-j",2)) {
         char *optlist = argv[argcount]+1;
         while (*optlist) {
           switch (*optlist) {
              case 'r': matchRange  = 1; break;
              case 'x': matchExon   = 1; break;
              case 'o': matchStrand = 1; break;
              case 'h': matchGraph  = 1; break;
              case 'j': matchSplice = 1; break;
              default : fatalf(Usage, argv[0]);
           }
           optlist++;
         }
      } else
         fatalf(Usage, argv[0]);

      argcount++;
   }

   fp1 = ckopen(argv[2], "r");
   fp2 = ckopen(argv[4], "r");

   lastGAid = -1;

   numasv1 = maxasv1 = numasv2 = maxasv2 = 0;
   asv1 = asv2 = NULL;

   st1 = loadASV(fp1, &asv1, &numasv1, &maxasv1, argv[1]);
   st2 = loadASV(fp2, &asv2, &numasv2, &maxasv2, argv[3]);

   while (st1 || st2) {

      numASVs1 = numASVs2 = 0;
      maxASVs1 = maxASVs2 = 0;
      ASVs1 = ASVs2 = NULL;

      /* determine the correct lastGAid */

      if (st1 && (!st2 || (asv1[0].gaId<=asv2[0].gaId))) {
         if (asv1[0].gaId >= lastGAid)
             lastGAid = asv1[0].gaId;
         else {
             fprintf(stderr,"gaId's in ASVs1 are out of order: %d<%d\n",
                     asv1[0].gaId, lastGAid);
             break;
         }
      } else if (st2 && (!st1 || (asv2[0].gaId<=asv1[0].gaId))) {
         if (asv2[0].gaId >= lastGAid)
             lastGAid = asv2[0].gaId;
         else {
             fprintf(stderr,"gaId's in ASVs2 are out of order: %d<%d\n",
                     asv2[0].gaId, lastGAid);
             break;
         }
      } else
         fatal("Something wrong when reading.\n");

      /* read all records for that lastGAid */
      while (st1 && (asv1[0].gaId==lastGAid)) {
         addMatches(asv1,numasv1,&ASVs1,&numASVs1,&maxASVs1); 

         /* should already have been transferred to the list in addMatch */
         for (i=0; i<numasv1; i++) {
             if (asv1[i].exons) free(asv1[i].exons);
             if (asv1[i].introns) free(asv1[i].introns);
             asv1[i].exons      = NULL;
             asv1[i].numExons   = 0;
             asv1[i].maxExons   = 0;
             asv1[i].introns    = NULL;
             asv1[i].numIntrons = 0;
             asv1[i].maxIntrons = 0;
         }
         numasv1 = 0;
    
         st1 = loadASV(fp1, &asv1, &numasv1, &maxasv1, argv[1]);
      }

      while (st2 && (asv2[0].gaId==lastGAid)) {
         addMatches(asv2,numasv2,&ASVs2,&numASVs2,&maxASVs2);

         /* should already have been transferred to the list in addMatch */
         for (i=0; i<numasv2; i++) {
             if (asv2[i].exons) free(asv2[i].exons);
             if (asv2[i].introns) free(asv2[i].introns);
             asv2[i].exons      = NULL;
             asv2[i].numExons   = 0;
             asv2[i].maxExons   = 0;
             asv2[i].introns    = NULL;
             asv2[i].numIntrons = 0;
             asv2[i].maxIntrons = 0;
         }
         numasv2 = 0;

         st2 = loadASV(fp2, &asv2, &numasv2, &maxasv2, argv[3]);
      }
      fprintf(stderr, "Computing for gaId=%d...\n", lastGAid);
      compare_asvs_sets(ASVs1,numASVs1,ASVs2,numASVs2,matchRange,matchExon,matchStrand,matchGraph,matchSplice,Vm,Vi,mode);

      /* free ASVs sets */
      for (i=0; i<numASVs1; i++) {
         if (ASVs1[i].exons) free(ASVs1[i].exons);
         if (ASVs1[i].introns) free(ASVs1[i].introns);
      }
      for (i=0; i<numASVs2; i++) {
         if (ASVs2[i].exons) free(ASVs2[i].exons);
         if (ASVs2[i].introns) free(ASVs2[i].introns);
      }
      numASVs1 = 0;
      numASVs2 = 0;
   }
   
   fclose(fp1); fclose(fp2);

   free(ASVs1); free(ASVs2);

   exit(0);
}

static int loadASV(FILE *fp, ASV_t **asvs, int *numasvs, int *maxasvs, char *flag)
{
   char   buf[100000], *st, rs; 
   ASV_t *asv;

   if (!strcmp(flag, "-t")) {
      /* read in a transcript */
      if (*numasvs==0) {
         *asvs = (ASV_t *)malloc(sizeof(ASV_t));
      }
      asv = (*asvs)+*numasvs;

      *(asv->asvId)   = '\0';
      asv->gaId       = -1;
      asv->graphId    = -1;
      asv->ccId       = -1;
      asv->strand     = -1;
      asv->relori     = -1;
      asv->b = asv->e = -1;
      asv->numExons   =  0;
      asv->maxExons   =  0;
      asv->numIntrons =  0;
      asv->maxIntrons =  0;
      asv->exons      = NULL;
      asv->introns    = NULL;
      asv->type       = TXPT_TYPE;

      st = fgets(buf, 100000, fp);
      rs = st ? line2ASV(buf, asv) : 0;             /* 0 is bad */

      if (rs) { (*numasvs)++; (*maxasvs)++; }

      return rs;
   } else if (!strcmp(flag, "-g")) {
      /* read in a graph */
      rs = readGraph(fp, asvs, numasvs, maxasvs);   /* 0 is bad */

      return rs;

   } else {
      sprintf(errmsg, "%c", flag);
      fatalf("Unrecognized flag (%s).\n", errmsg);
   }
}


static int line2ASV(char *buf, ASV_t *p)
{
   char *tok;
   int   nodeid, from, to, i;

   /* ASV_2586 0 98700007.0 -1 -1 ND_98703668(99668261-99669032) ND_98703669(99669056-99669247) */

   if (*buf=='\0') return 0;

   tok = strtok(buf, " \n"); 
   strcpy(p->asvId,tok);

   tok = strtok(NULL, " \n");
   p->gaId = atoi(tok);

   tok = strtok(NULL, " \n");
   if (sscanf(tok, "%d.%d", &(p->graphId), &(p->ccId))!=2)
      fatalf("GraphId token expected. %s\n", buf);

   tok = strtok(NULL, " \n");
   p->strand = atoi(tok);
   if (p->strand!=1 && p->strand!=-1) 
       fatalf("Unexpected value for ASV strand. %s\n", tok);

   tok = strtok(NULL, " \n");
   p->relori = atoi(tok);
   if (p->relori!=1 && p->relori!=-1) 
       fatalf("Unexpected value for ASV relori. %s\n", tok);

   tok = strtok(NULL, " \n");
   while (tok) {
      Exon_t *x;
        
      if (sscanf(tok, "ND_%d(%d-%d)", &nodeid, &from, &to)!=3)
         fatalf("Node token expected. %s\n", tok); 

      if (p->numExons>=p->maxExons) {
         p->maxExons = max(2*p->maxExons, DEFAULT_ALLOC_SMALL);
         p->exons    = realloc(p->exons, p->maxExons*sizeof(Exon_t));
      }
      x = p->exons+p->numExons;

      x->nodeId = nodeid; 
      x->from = from;
      x->to = to;

      p->numExons++;

      tok = strtok(NULL, " \n");
   }

   /* now infer the introns */
   p->introns = (p->numExons==1) ? NULL :
                (Intron_t *)malloc((p->numExons-1)*sizeof(Intron_t));
   p->numIntrons = p->maxIntrons = p->numExons-1;

   for (i=0; i<p->numExons-1; i++) {
      /* no edgeId yet */
      p->introns[i].from = p->exons[i].to;
      p->introns[i].to   = p->exons[i+1].from;
   }

   p->type = TXPT_TYPE;
   p->b    = p->exons[0].from;
   p->e    = p->exons[p->numExons-1].to;

   return 1;
}

static int readGraph(FILE *fp, ASV_t **genes, int *numgenes, int *maxgenes)
{
   char   *buf, *tok;
   int     bufsize = 0, i, j;
   char   *st;
   ASV_t  *g;


   buf = NULL; bufsize = 0;

   while ((st=getLine(&buf,&bufsize,fp))!=NULL) {

      if (!strncmp(buf, "||", 2)) break;     /* end of graph */

      if (*buf == '>') {

         /* create another graph */
         if (*numgenes>=*maxgenes) {
            *maxgenes = max(2*(*maxgenes), DEFAULT_ALLOC_SMALL);
            *genes = realloc(*genes, *maxgenes*sizeof(ASV_t));
         }
         g = (*genes)+(*numgenes);

         tok = strtok(buf, " ");
         sscanf(tok+4, "%d.%d", &(g->graphId), &(g->ccId));
         strcpy(g->asvId, tok+4);

         tok = strtok(NULL, " ");
         sscanf(tok, "%d[", &(g->gaId));

         tok = strtok(NULL, " ");
         if (tok) {
            if (!strncmp(tok, "<forward-", 9))
                g->strand = g->relori = 1;
            else if (!strncmp(tok, "<reverse-", 9))
                g->strand = g->relori = -1;
            else
                fatalf("Unexpected strand assignment in GS line (%s).", tok);
         }

         (*numgenes)++;

         getLine(&buf,&bufsize,fp);          /* skip over EST evidence */

      } else if (!strncmp(buf, "ND", 2)) {

         Exon_t      *x;
         char         graphid[15];
         long int     nodeid;
         int          gidx, start, end;

         tok = strtok(buf, " ");
         nodeid = atol(tok+3);    /* LONG INT */

         tok = strtok(NULL, " ");
         start = atoi(tok);

         tok = strtok(NULL, " ");
         end = atoi(tok);

         tok = strtok(NULL, " ");
         tok = strtok(NULL, " ");
         strcpy(graphid, tok+3);

         for (i=0, gidx = -1; i<*numgenes && gidx<0; i++)
            if (!strcmp(graphid, (*genes)[i].asvId)) gidx = i;
         if (gidx<0) fatalf("Could not find graph %s.\n", graphid);

         g = (*genes) + gidx;

         if (g->numExons == g->maxExons) {
             g->maxExons =  max(DEFAULT_ALLOC_SMALL, 2 * g->maxExons);
             g->exons    =  realloc(g->exons, g->maxExons*sizeof(Exon_t));
         }

         x = g->exons + g->numExons;

         x->nodeId = nodeid;
         x->from   = start;
         x->to     = end;

         g->numExons++;

      } else if (!strncmp(buf, "EG", 2)) {

           Intron_t  *eg;
           long int   nodeid1, nodeid2;
           int        i, j, gidx, start, end;


           tok = strtok(buf, " ");   /* EG */
           tok = strtok(NULL, " ");

           nodeid1 = atol(tok+3);

           for (i=0, gidx=-1; i<*numgenes; i++)
              for (j=0; j<(*genes)[i].numExons; j++)
                 if ((*genes)[i].exons[j].nodeId == nodeid1) {
                    gidx = i;
                    start = (*genes)[i].exons[j].to;
                    break;
                 }

           if (gidx < 0) {
              sprintf(errmsg, "%d", nodeid1);
              fatalf("Could not find node %s.\n",  errmsg);
           }
           g = (*genes) + gidx;

           tok = strtok(NULL, " ");

           while (tok != NULL) {
                sscanf(tok, "ND_%ld:", &nodeid2);

                for (j=0, end=-1; j<g->numExons; j++)
                   if (g->exons[j].nodeId == nodeid1) {
                      end = g->exons[j].from;
                      break;
                   }

                if (end<0) {
                    sprintf(errmsg, "%d", nodeid2); 
                    fatalf("Could not find node %s.\n",  errmsg);
                }

                if (g->numIntrons == g->maxIntrons) {
                    g->maxIntrons =  max(g->maxIntrons * 2, DEFAULT_ALLOC_SMALL);
                    g->introns    =  realloc(g->introns, g->maxIntrons*sizeof(Intron_t));
                }

                eg = g->introns + g->numIntrons;
                eg->from = start;
                eg->to   = end;
                g->numIntrons++;

                tok = strtok(NULL, " ");
           }

      } else if (!strncmp(buf, "AS", 2)) {
         ;
      } else if (!strncmp(buf, "XS", 2)) {
         ;
      } else if (!strncmp(buf, "Cluster", 7)) {
         ;
      }
   }

   for (i=0; i<*numgenes; i++) {
      g = (*genes)+i;
      g->b = g->exons[0].from;
      g->e = g->exons[0].to;
      for (j=1; j<g->numExons; j++) {
         g->b = min(g->b, g->exons[j].from);
         g->e = max(g->e, g->exons[j].to);
      }
   }

   if (buf) free(buf);
   return (st==NULL) ? 0 : 1;
}


static void compare_asvs_sets(ASV_t *ASVs1, int numASVs1, ASV_t *ASVs2, int numASVs2, int mrange, int mexon, int mstrand, int mgraph, int msplice, int Vm, int Vi, int mode)
{
   int     i, j;
   ASV_t  *ai, *aj;
   int     st;

   CompareStats_t  es, et;
 

   for (i=0; i<numASVs1; i++) {
     ai = ASVs1+i;
     for (j=0; j<numASVs2; j++) {
       aj = ASVs2+j;  

       /* establish whether the ASV can be compared with this XSV */
       if ((mrange && !overlaps(ai->b, ai->e, aj->b, aj->e)) ||
           (mexon && !exon_overlaps(ai,aj)) ||
           (mstrand && (ai->strand != aj->strand)) ||
           (mgraph && (ai->graphId!=aj->graphId)) ||
           (msplice && (ai->numExons>1) && (aj->numExons>1) && !common_splice(ai, aj)))

           continue;

/*     printf("<<<%s (%d %d) %s (%d %d)\n", ai->asvId, ai->b, ai->e, aj->asvId, aj->b, aj->e); */
       st = feature_compare(ai, aj, &es, &et, Vi, Vm, mode);
     }
   }

   return;
}

static int common_splice(ASV_t *ai, ASV_t *aj)
{
    int i, j;

    for (i=0; i<ai->numIntrons; i++) {
       int sl = ai->introns[i].from;
       int sr = ai->introns[i].to;

       for (j=0; j<aj->numIntrons; j++) 
         if (abs(aj->introns[j].from-sl)<=V || abs(aj->introns[j].to-sr)<=V)
             return 1;
    }
    return 0;
}

/* transfer the exons from asv to the list; make sure to reset them to NULL */
static void addMatches(ASV_t *newasvs, int numnewasvs, ASV_t **asvsptr, int *numasvs, int *maxasvs)
{
    ASV_t  *asv, *in_asv;
    int     i;

    if (*numasvs+numnewasvs>=*maxasvs) {
        *maxasvs = max(2*(numnewasvs+*maxasvs), DEFAULT_ALLOC_LARGE); 
        *asvsptr = realloc(*asvsptr, *maxasvs*sizeof(ASV_t)); 
    }

    for (i=0; i<numnewasvs; i++) {
        asv = (*asvsptr) + (*numasvs);
        in_asv = newasvs + i;

        strcpy(asv->asvId, in_asv->asvId);

        asv->gaId       = in_asv->gaId;
        asv->graphId    = in_asv->graphId;
        asv->ccId       = in_asv->ccId;
        asv->strand     = in_asv->strand;
        asv->relori     = in_asv->relori;
        asv->numExons   = in_asv->numExons; 
        asv->maxExons   = in_asv->maxExons; 
        asv->exons      = in_asv->exons;
        asv->numIntrons = in_asv->numIntrons;
        asv->maxIntrons = in_asv->maxIntrons;
        asv->introns    = in_asv->introns;
        asv->b          = in_asv->b;
        asv->e          = in_asv->e;
        asv->type       = in_asv->type;

        (*numasvs)++;

        /* leave numExons unchanged -- why? LLL */
        in_asv->exons = NULL;
        in_asv->maxExons = 0;
        in_asv->introns = NULL;
        in_asv->maxIntrons = 0;
    }
}

/* ----------------------   Line mng utilities ------------------ */

#define DEFAULT_BUF_SIZE  100000

static char *getLine(char **buf, int *bufsize, FILE *fp)
{
   char *end, *p;
   int   c, oldbufsize;

   if (*buf == NULL) {
       *buf = (char *)malloc(DEFAULT_BUF_SIZE*sizeof(char));
       *bufsize = DEFAULT_BUF_SIZE;
   }

   p = *buf; *p = '\0'; end = *buf + *bufsize -1;  /* allow space for the string terminator */

   while ((c=fgetc(fp))!=EOF && c!='\n') {
     if (p<end) *p++ = c;
     else {
         oldbufsize = *bufsize;
        *bufsize = 2*(*bufsize);
        *buf = realloc(*buf,*bufsize);
         end = *buf + *bufsize -1;
         p = *buf + oldbufsize-1;
        *p++ = c;
     }
   }
   *p = '\0';

   if (c==EOF && *buf[0]) fputc(c,fp);
   else if (c==EOF) return NULL;

   return *buf;
}

static int exon_overlaps(ASV_t *a, ASV_t *b) 
{
  int i, j;

  for (i=0; i<a->numExons; i++) 
     for (j=0; j<b->numExons; j++) 
         if (overlaps(a->exons[i].from, a->exons[i].to, b->exons[j].from, b->exons[j].to))
            return 1;
  return 0;
}
   
